#!/usr/bin/env python

##  Copyright (C) 1998  marduk <marduk@starship.skyport.net>

##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or
##  (at your option) any later version.

##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##  GNU General Public License for more details.

##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


__version__='0.9'
__author__ = 'marduk (marduk@starship.skyport.net)'

USAGE='linbot [-abv][-x url]... [-y url]... [-o dir][-w sec] url [location]...'
PYTHON_VERSION=1.5 # not used right now
level = 0
explored = []
problem_db = {}
linkList = {}

import sys
import time
import string

start_time = time.ctime(time.time())

# importing the config.py file is a real problem if the user did not install
# the files EXACTLY the way I said to... or even using the frozen version is
# becoming a real bitch.  I will just have to tell them right out how to fix it.
try:
    sys.path = ['.'] + sys.path
    import config
except ImportError:
    sys.stdout.write('Please verify that PYTHONPATH knows where to find "config.py"\n')
    sys.exit(1)

import myUrlLib
Link=myUrlLib.Link

# myUrlLib will be looking for a 'config' module.  set it up here.
myUrlLib.config=config

CODES=myUrlLib.HTTP_STATUS_CODES

def parse_args():
    import getopt
    global URL
    try:
	optlist, args = getopt.getopt(sys.argv[1:],'vx:y:ao:bw:')
    except getopt.error, reason:
	print reason
	print USAGE
	sys.exit(1)
    for flag,arg in optlist:
	if flag=='-v':
	    print_version()
	    sys.exit(0)
	elif flag=='-x':
	    config.EXCLUDED_URLS.append(myUrlLib.urlformat(arg))
	elif flag=='-y':
	    config.YANKED_URLS.append(myUrlLib.urlformat(arg))
	elif flag=='-a':
	    config.AVOID_EXTERNAL_LINKS=1
	elif flag=='-o':
	    config.OUTPUT_DIR=arg
	elif flag=='-b':
	    config.BASE_URLS_ONLY=1
	elif flag=='-w':
	    config.WAIT_BETWEEN_REQUESTS=int(arg)

    if len(args)==0:
	print USAGE
	sys.exit(1)
    else: URL = args[0]
    config.HOSTS=args[1:]

def sort_by_age(a,b):
    """ sort helper for url's age.  a and b are urls in linkList """
    aage, bage = linkList[a].age, linkList[b].age
    if aage < bage:
	return -1
    if aage == bage:
	return sort_by_author(a,b)
    return 1

def sort_by_rev_age(a,b):
    aage, bage = linkList[a].age, linkList[b].age
    if aage > bage:
	return -1
    if aage == bage:
	return sort_by_author(a,b)
    return 1

def sort_by_author(a,b):
    aauthor,bauthor = `linkList[a].author`, `linkList[b].author`
    if aauthor < bauthor:
	return -1
    if aauthor == bauthor:
	return 0
    return 1

def sort_by_size(a,b):
    asize, bsize = linkList[a].totalSize, linkList[b].totalSize
    if asize < bsize:
	return 1
    if asize == bsize:
	return 0
    return -1

def open_file(filename):
    """ given filename (and config.OUTPUT_DIR) opens filename for writing and
    returns the file object """
    return open(config.OUTPUT_DIR + filename,'w')
    
def get_title(url):
    """ returns the title of a url if it is not None, else returns url
    note that this implies linkList[url] """
    link=linkList[url]
    if link.title is None:
	return url
    return link.title

def make_link(url,text):
    """Return an <A>nchor to a url with <text>.  If url is in the Linklist and
    is external, insert "class=external" in the <A> tag."""
    url = str(url) # because sometimes I lazily pass a Link object.
    mystring = '<a href="' + url + '"'
    try:
	external = linkList[url].external
    except KeyError:
	external = 0
    if external:
	mystring = mystring + ' class="external"'
    else:
	mystring = mystring + ' class="internal"'
    mystring = mystring + '>' + text + '</a>'
    return mystring
    

def add_problem(type,link):
    """ add a problem to the 'problems' database.  Will not add external links"""
    if link.external: return
    global problem_db
    author = link.author
    if problem_db.has_key(author):
	problem_db[author].append((type,link))
    else:
	problem_db[author]=[(type,link)]

class Page:
    # this is a class which describes an HTML pointed to in the navbar
     def __init__(self,title,filename,function,image=None):
	self.title = title
	self.filename = filename
	self.function = function
	self.image = image

	
def doTopMain(title='',function):
    """top part of html files in main frame prints to stdout"""
    print '<html>'
    print '<head><title>%s</title>' % title
    print '<link rel="stylesheet" type="text/css" href="%s">' \
	  % config.STYLESHEET
    print '</head>'
    print '<body class="%s">' % function.__name__  
    print '<p class="logo"><img src="%(LOGO_HREF)s" alt="%(LOGO_ALT)s">' % vars(config)
    print '\n<h1 class="basename">'
    print '\t<a href="%s">%s</a>' \
	  % (`base`, get_title(`base`))
    print '</h1>'
    print '\n\n<table width="100%" cellpadding=4>'
    print '\t<tr><th class="title">%s</th></tr>\n</table>\n' % title



def doBotMain():
    """ bottom part of html files in main frame"""
    print 
    print '<hr>'
    print '<p class="footer">'
    print '<em>Generated %s by <a target="_top" href="%s">Linbot %s</a></em>' \
	  % (start_time,config.LINBOT_HOME, __version__)
    print '</body>'
    print '</html>'


def main_index():
    fp = open_file(config.MAIN_FILENAME)
    sys.stdout=fp
    
    print '<html>'
    print '<head>'
    print '<title>Linbot report for "%s"</title>' % get_title(`base`)
    print '<link rel="stylesheet" type="text/css" href="%s">' \
	  % config.STYLESHEET
    print '</head>'
    print '<frameset COLS="%s,*" border=0 framespacing=0>' \
	  % config.NAVBAR_WIDTH
    print '<frame name="navbar" src="%s" marginwidth=0 marginheight=0>' \
	  % config.NAVBAR_FILENAME
    print '<frame name="main" src="%s">' % config.SITE_MAP_FILENAME
    print '</frameset>'
    print '</html>'
    fp.close()


def nav_bar(pages):
    # navigation bar
    fp=open_file(config.NAVBAR_FILENAME)
    sys.stdout=fp
    print '<html>\n<head>'
    print '\t<title>navbar</title>'
    print '\t<link rel="stylesheet" type="text/css" href="%s">' \
	  % config.STYLESHEET
    print '\t<base target="main">'
    print '</head>'
    print '<body class="navbar">'
    print '<div align=center>'
    print '<table width="100%%" cellpadding="%s" cellspacing="%s">' \
	  % (config.NAVBAR_PADDING, config.NAVBAR_SPACING)
    # title
    print '<tr><th class="home">',
    if config.LINBOT_IMAGE is None:
	print '<a target="_top" href="%s">Linbot %s</a></th></tr>' \
	      % (config.LINBOT_HOME, __version__)
    else:
	print '<A target="_top" href="%s"><img border=0 SRC="%s" alt="Linbot %s"></A></th></tr>' \
	      % (config.LINBOT_HOME, config.LINBOT_IMAGE, __version__)

    # 'buttons' pointing to each individual page
    for page in pages:
	print '<tr><th>',
	if page.image is None:
	    print '<strong><a href="%s">%s</a></strong>' \
		  % (page.filename, page.title),
	else:
	    print '<a href="%s"><img border=0 src="%s" alt="%s"></a>' \
		  % (page.filename, page.image, page.title),
	print '</th></tr>'

	# create the file we just pointed to
	tmp = sys.stdout
	fp = open_file(page.filename)
	sys.stdout = fp
	doTopMain(page.title,page.function)
	apply(page.function)
	doBotMain()
	fp.close()
	sys.stdout = tmp
	
    print
    print '</table>'
    print '</div>'
    print '</body>'
    print '</html>'

    fp.close()

def explore(link, explored):
    """Recursively do a breadth-first traversal of the graph of links
    on the site.  Returns a list of HTML fragments that can be printed 
    to produce a site map."""

    global level
    if level>config.SITE_MAP_LEVEL: return []
    # XXX I assume an object without a .URL is something
    # uninteresting? --amk
    if not hasattr(link, 'URL'): return []

    level=level+1
    explored[ link.URL ] = 1
    to_explore = []
    L = ['<ul>']

    # We need to do a breadth-first traversal.  This requires two
    # steps for any given page.  First, we need to make a list of
    # links to be traversed; links that have already been explored can 
    # be ignored.
    
    for i in link.children:
        # Skip pages that have already been traversed
        if explored.has_key( i ): continue
	if (i in Link.badLinks) and not config.ANCHOR_BAD_LINKS:
	    L.append('<li>%s' % i)
	else:
            to_explore.append(i)
        explored[ i ] = 1               # Mark the link as explored

    # Now we loop over the list of links; the traversal will not go to 
    # any pages that are marked as having already been traversed.
    for i in to_explore:
            child = linkList[i]
	    L.append('<li>%s' % (make_link(i,get_title(i))))
	    L = L + explore(child, explored)
            
    L.append( '</ul>' )
    level=level-1

    # If no sub-pages were traversed at all, just return an empty list 
    # to avoid redundant <UL>...</UL> pairs
    if len(L) == 2: return []

    return L
    
# site map
def site_map():	
    print make_link(base,'Starting Page')
    L = explore(base, {})
    for i in L: print i

#bad links
def bad_links():
    print '<div class="table">'
    print '<table border=0 cellspacing=2 width="75%">'
    for link in Link.badLinks:
	print '\t<tr><td class="blank" colspan=3>&nbsp;</td></tr>'
	if config.ANCHOR_BAD_LINKS:
	    print '\t<tr class="link"><th>Link</th>',
	    print '<td colspan=2 align=left>'  +make_link(link,link) +'</td></tr>'
	else:
	    print '\t<tr class="link"><th>Link</th>',
	    print '<td colspan=2 align=left>%s</td></tr>' % link
	status = str(linkList[link].status)
	if status in CODES.keys():
	    status = status + ": " + CODES[status]
	print '\t<tr class="status"><th>Status</th><td colspan=2>%s</td></tr>' % status
	print '\t<tr class="parent"><th rowspan="%s">Parents</th>' % len(linkList[link].parents)
	parents = linkList[link].parents
	parents.sort(sort_by_author)
	for parent in parents:
	    print '\t\t<td>%s</td>' % make_link(parent,get_title(parent)),
	    print '<td>%s</td>\n\t</tr>' % (str(linkList[parent].author))
	    add_problem("Bad Link: " + link,linkList[parent])
    print '</table>'
    print '</div>'


# images
def images():    
    import math
    imagelist=Link.images.keys()

    currentPic=0
    rows = int(math.ceil(len(imagelist)/config.IMAGE_COLS))+1
    print '<div class="table">'
    print '<table border=0>'
    
    for row in range(rows):
	print'\t<tr>'
	for col in range(config.IMAGE_COLS):
	    if currentPic==len(imagelist): break
	    image=imagelist[currentPic]
	    print '\t\t<td>' + \
	      make_link(image,
			'<img src="%s" width="%d" height="%d" alt="%s">' \
		  % (image,config.IMAGE_WIDTH,config.IMAGE_HEIGHT, image)),
	    print '</td>'
	    currentPic = currentPic + 1

	print '\t</tr>'
    print '</table>'
    print '</div>'

# what's new
def whats_new():
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    print '\t<tr><th>Link</th><th>Author</th><th>Age</th></tr>'
    urls = linkList.keys()
    urls.sort(sort_by_age)
    for url in urls:
	link=linkList[url]
	if not link.html: continue
	age = link.age
	if (age is not None)and (age <= config.NEW_URL_AGE):
	    print '\t<tr><td>%s</td>' % make_link(url,get_title(url)),
	    print '<td>%s</td>' % link.author,
	    print '<td class="time">%s</td></tr>' % age
    print '</table>'
    print '</div>'

# what's old
def whats_old():
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    print '\t<tr><th>Link</th><th>Author</th><th>Age</th></tr>'
    urls = linkList.keys()
    urls.sort(sort_by_rev_age)
    for url in urls:
	link=linkList[url]
	if not link.html: continue
	age = link.age
	if age and (age >= config.OLD_URL_AGE):
	    print '\t<tr><td>%s</td>' % make_link(url,get_title(url)),
	    print '<td>%s</td>' % (link.author),
	    print '<td class="time">%s</td></tr>' % age
	    add_problem('Old Link: %s days old' % age ,link)
    print '</table>'
    print '</div>'

# what's slow
def whats_slow():
    import time
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    print '\t<tr><th rowspan=2>Link</th>',
    print '<th rowspan=2>Size <br>(Kb)</th>',
    print '<th colspan=3>Time (HH:MM:SS)</th></tr>'
    print '\t<tr><th>28.8</th><th>ISDN</th><th>T1</th></tr>'

    urls = linkList.keys()
    urls.sort(sort_by_size)
    for url in urls:
	link = linkList[url]
	if not link.html: continue
	sizeK = link.totalSize / 1024
	sizek = link.totalSize * 8 / 1000
	if sizeK < config.SLOW_URL_SIZE:
	    break
	print '\t<tr><td>%s</td>' % make_link(url, get_title(url)),
	print '<td>%s</td><td class="time">%s</td>' \
	      % (sizeK, time.strftime('%H:%M:%S',time.gmtime(int(sizek/28.8)))),
	print '<td class="time">%s</td>' \
	      % time.strftime('%H:%M:%S',time.gmtime(int(sizek/56))),
	print '<td class="time">%s</td>' \
	      % time.strftime('%H:%M:%S',time.gmtime(int(sizek/1500))),
	print '</tr>'
	add_problem('Slow Link: %sK' % sizeK, link) 
    print '</table>'
    print '</div>'

# pages with no titles
def no_titles():
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    print '\t<tr><th>URL</th><th>Author</th></tr>'
    urls = linkList.keys()
    urls.sort(sort_by_author)
    for url in urls:
	link = linkList[url]
	if link.external: continue
	if link.html and (link.title is None):
	    print '\t<tr><td>%s</td><td>%s</td></tr>' \
		  % (make_link(url,url), link.author)
	    add_problem("No Title",link)
    print '</table>'
    print '</div>'
	
# pages which were not checked
def not_checked():
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    for url in Link.notChecked:
	print '\t<tr><th colspan=4>%s</th></tr>' % make_link(url,url)
	print '\t<tr class="parent"><th rowspan="%s">Parent</th>' % len(linkList[url].parents)
	for parent in linkList[url].parents:
	    print '\t\t',
	    if parent != linkList[url].parents[0]: print '<tr>',
	    print '<td colspan=2>%s</td>' % make_link(parent,get_title(parent)),
	    print '<td>%s</td></tr>' % (linkList[parent].author)
	print '\n\t<tr><td class="blank" colspan=4>&nbsp;</td></tr>\n'
    print '</table>'
    print '</div>'

# external links
def external():
    print '<ol>'
    for url in linkList.keys():
	link=linkList[url]
	if link.external:
	    print '\t<li>%s' % make_link(url,get_title(url))
    print '</ol>'

# breakdown of links with problems
def problems():
    authors=problem_db.keys()
    authors.sort()
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    for author in authors:
	print '<tr><th>%s</th></tr>' % author
	for type,link in problem_db[author]:
	    url=`link`
	    title=get_title(url)
	    print '<tr><td>%s ( %s )</td></tr>' % (make_link(url,title), type)
	print '<tr><td class="blank">&nbsp;</td></tr>\n'
    print '</table>'
    print '</div>'

# configuration options
def config_page():
    print '<div class="table">'
    print '<table border=0 cellpadding=2 cellspacing=2 width="75%">'
    print '<tr><th>Name</th><th>Value</th></tr>\n'
    variables = vars(config)
    sorted=variables.keys()
    sorted.sort()
    for variable in sorted:
	if variable[:2]=="__": continue # don't show special variables
	print '<tr><td>%s</td>' % variable,
	print '<td>%s</td></tr>' % variables[variable]
    print '<tr><td>Command</td><td>%s</td></tr>' % string.join(sys.argv)
    print '</table>'
    print '</div>'

def get_version(): # broken... DO NOT USE
    import string
    version = string.split(sys.version)[0]
    return eval(version)

def print_version():
    """Print version information"""
    import os
    print "    Linbot: " + __version__
    print "    Python: " + sys.version
    print "    OS:     " + os.name
    print

def warn():
    """Warn the user that something has gone wrong."""
    print "*******************************************"
    print "*                                         *"
    print "* Warning, Linbot has found nothing to    *"
    print "* report for this site.  If you feel this *"
    print "* is in error, please contact             *"
    print "* %s.                                     *" % __author__
    print "* and specify the environment that caused *"
    print "* this to occur.                          *"
    print "*                                         *"
    print "* Linbot %s                               *" % __version__
    print "*                                         *"
    print "*******************************************"

if __name__=='__main__':

    # make sure we are running correct version of python
    # get_version() is broken (it doesn't work for, say 1.5.1) so we will
    # for now just skip version checking.
##     if get_version() < PYTHON_VERSION:
## 	sys.stderr.write("Linbot must be run using Python %s or better\n" \
## 			 % PYTHON_VERSION)
## 	sys.exit(1)

    parse_args()
    
    config.OUTPUT_DIR=config.OUTPUT_DIR + '/'

    
    print 'checking site....'
    base=myUrlLib.Link(URL,None) # this will take a while
    print 'done.'
    if not hasattr(base,"URL"):
	warn()
	sys.exit(1)

    linkList = Link.linkList

    # now we can write out the files
    # start with the frame-description page
    main_index()
    
    # set up the pages
    pages = []
    pages.append(Page(config.SITE_MAP_TITLE,
		      config.SITE_MAP_FILENAME,
		      site_map,
		      config.SITE_MAP_IMAGE))
    pages.append(Page(config.BAD_LINKS_TITLE,
		      config.BAD_LINKS_FILENAME,
		      bad_links,
		      config.BAD_LINKS_IMAGE))
    pages.append(Page(config.IMAGE_TITLE,
		      config.IMAGE_FILENAME,
		      images,
		      config.IMAGE_IMAGE))
    pages.append(Page(config.OLD_URL_TITLE,
		      config.OLD_URL_FILENAME,
		      whats_old,
		      config.OLD_URL_IMAGE))
    pages.append(Page(config.NEW_URL_TITLE,
		      config.NEW_URL_FILENAME,
		      whats_new,
		      config.NEW_URL_IMAGE))
    pages.append(Page(config.SLOW_URL_TITLE,
		      config.SLOW_URL_FILENAME,
		      whats_slow,
		      config.SLOW_URL_IMAGE))
    pages.append(Page(config.NO_TITLES_TITLE,
		      config.NO_TITLES_FILENAME,
		      no_titles,
		      config.NO_TITLES_IMAGE))
    pages.append(Page(config.EXTERNAL_TITLE,
		      config.EXTERNAL_FILENAME,
		      external,
		      config.EXTERNAL_IMAGE))
    pages.append(Page(config.NOT_CHECKED_TITLE,
		      config.NOT_CHECKED_FILENAME,
		      not_checked,
		      config.NOT_CHECKED_IMAGE))
    pages.append(Page(config.PROBLEMS_TITLE,
		      config.PROBLEMS_FILENAME,
		      problems,
		      config.PROBLEMS_IMAGE))
    if config.SHOW_CONFIG_PAGE:
	pages.append(Page(config.CONFIG_TITLE,
			  config.CONFIG_FILENAME,
			  config_page,
			  config.CONFIG_IMAGE))
    
    # create the navigation bar, which creates the pages it points to
    nav_bar(pages)









 

    

